#include <string>
#include <iostream>
#include <cassert>

#include "lmxsoap.h"
#include "lmxwinhttp.h"
#include "lmxparse.h"

#include "soap-example-generated.h"

const char * your_name = { "Bob" }; // Set to your name

void manually_marshalled_soap_example();
void manually_marshalled_soap_with_fault_example();
void object_to_object_soap_example();
void rest_example();

int main()
{
    // This file shows four examples of how to use LMX's SOAP/WS functionality.

    manually_marshalled_soap_example();

    manually_marshalled_soap_with_fault_example();

    object_to_object_soap_example();

    rest_example();

    std::cout << "\n\n--Hit Return to exit--\n";
    std::cin.peek();

    return 0;
}

void manually_marshalled_soap_example()
{
    // This example shows separately creating SOAP body objects, wrapping
    // them in the SOAP wrapper object, and then marshalling the result
    // to a std::string.  The message is then sent to the server, the
    // response obtained and displayed.

    std::cout << "In manually_marshalled_soap_example()\n";

    // Create the appropriate SOAP wrapper object and add body and header objects
    lmx::c_soap< SOAPExample::c_Request, SOAPExample::c_Header, SOAPExample::c_Faults >
            soap_request;

    std::auto_ptr< SOAPExample::c_Request > ap_request( new SOAPExample::c_Request );
    ap_request->set_name( lmx::convert( your_name ) );
    soap_request.bodies().push_back( ap_request.get() );     // The SOAP wrapper will delete the pointer when it is finished with
    ap_request.release();

    std::auto_ptr< SOAPExample::c_Header > ap_header( new SOAPExample::c_Header );
    ap_header->set_maxHops( 1 );
    soap_request.headers().push_back( ap_header.get() );     // The SOAP wrapper will delete the pointer when it is finished with
    ap_header.release();

    // Create the SOAP XML message from the objects
    std::string soap_request_message;
    lmx::elmx_error error = soap_request.marshal( &soap_request_message );

    if( error == lmx::ELMX_OK )
    {
        // Connect to the server, send the message and get the reply
        lmx::c_winhttp http( "codalogic.com" );

        std::string soap_response_message;
        int http_error = http.soap_rpc(
                                "services/",                // URI path
                                L"SOAPExample",             // SOAPAction
                                soap_request_message,     // Request message
                                &soap_response_message ); // Response message

        if( http_error != 0 )
            std::cout << "HTTP error occured: " << lmx::c_winhttp::get_error_description( http_error ) << "\n";

        else
        {
            // Create objects from the reply message
            lmx::elmx_error lmx_error;
            const lmx::c_soap< SOAPExample::c_Response, SOAPExample::c_Header, SOAPExample::c_Faults >
                    soap_response( soap_response_message, &lmx_error );

            if( lmx_error != lmx::ELMX_OK )
                std::cout << "LMX error occurred: " << lmx::get_error_description( lmx_error ) << "\n";

            else
            {
                if( ! soap_response.bodies().empty() )
                {
                    const SOAPExample::c_Response *p_response = soap_response.bodies()[0];
                    std::wcout << L"Message: " << p_response->get_message() << L"\n";
                    std::wcout << L"Host: " << p_response->get_hostName() << L"\n";
                    std::cout << "Datetime: " << p_response->get_datetime() << "\n";
                }
            }
        }
    }
}

void manually_marshalled_soap_with_fault_example()
{
    // This example is the same as the manually_marshalled_soap_example(), but calls
    // the wrong SOAPAction and therefore receives a SOAP fault in reply.

    std::cout << "\nIn manually_marshalled_soap_with_fault_example()\n";

    // Create the appropriate SOAP wrapper object and add body and header objects
    lmx::c_soap< SOAPExample::c_Request, SOAPExample::c_Header, SOAPExample::c_Faults >
            soap_request;

    std::auto_ptr< SOAPExample::c_Request > ap_request( new SOAPExample::c_Request );
    ap_request->set_name( lmx::convert( your_name ) );
    soap_request.bodies().push_back( ap_request.get() );     // The SOAP wrapper will delete the pointer when it is finished with
    ap_request.release();

    std::auto_ptr< SOAPExample::c_Header > ap_header( new SOAPExample::c_Header );
    ap_header->set_maxHops( 1 );
    soap_request.headers().push_back( ap_header.get() );     // The SOAP wrapper will delete the pointer when it is finished with
    ap_header.release();

    // Create the SOAP XML message from the objects
    std::string soap_request_message;
    lmx::elmx_error error = soap_request.marshal( &soap_request_message );

    if( error == lmx::ELMX_OK )
    {
        // Connect to the server, send the message and get the reply
        lmx::c_winhttp http( "codalogic.com" );

        std::string soap_response_message;
        int http_error = http.soap_rpc(
                                "services/",                // URI path
                                L"SOAPFault",               // SOAPAction - in this case, invalid
                                soap_request_message,     // Request message
                                &soap_response_message ); // Response message

        if( http_error != 0 )
            std::cout << "HTTP error occured: " << lmx::c_winhttp::get_error_description( http_error ) << "\n";

        else
        {
            // Create objects from the reply message
            lmx::elmx_error lmx_error;
            const lmx::c_soap< SOAPExample::c_Response, SOAPExample::c_Header, SOAPExample::c_Faults >
                    soap_response( soap_response_message, &lmx_error );

            if( lmx_error != lmx::ELMX_OK )
            {
                MK_LMX_DEBUG_ERROR_PTR( p_debug_error );       // Provides easier access to LMX debug information when debugging
                std::cout << "LMX error occurred: " << lmx::get_error_description( lmx_error ) << "\n";
            }

            else
            {
                if( ! soap_response.bodies().empty() )
                {
                    // We shouldn't go down this path
                    assert( 0 );

                    const SOAPExample::c_Response *p_response = soap_response.bodies()[0];
                    std::wcout << L"Message: " << p_response->get_message() << L"\n";
                    std::wcout << L"Host: " << p_response->get_hostName() << L"\n";
                    std::cout << "Datetime: " << p_response->get_datetime() << "\n";
                }

                else if( soap_response.is_fault() )
                {
                    std::cout << "Error code (namespace): " <<
                            soap_response.get_fault().get_fault_code().get_namespace() << std::endl;
                    std::wcout << L"Error code (local part): " <<
                            soap_response.get_fault().get_fault_code().get_local_name() << L"\n";
                    std::wcout << L"Error string: " << soap_response.get_fault().get_fault_string() << L"\n";
                    if( soap_response.get_fault().isset_details() )
                        std::wcout << L"Allowed SOAPActions: " <<
                                soap_response.get_fault().details()[0]->get_allowedSOAPActions() << std::endl;
                }
            }
        }
    }
}

void object_to_object_soap_example()
{
    // This example shows the simple SOAP interface.  This only allows
    // the SOAP exchange to consist of body parts.  However, the interface
    // is templatized so that an object can be input to the function and
    // an object is returned, thus greatly simplifying the interface to
    // the service.

    std::cout << "\nIn object_to_object_soap_example()\n";

    SOAPExample::c_Request request;

    request.set_name( lmx::convert( your_name ) );

    SOAPExample::c_Response *p_response;

    lmx::s_winhttp_error result =
            lmx::simple_soap(
                    "codalogic.com",    // Host
                    "services/",        // URI path
                    L"SOAPExample",     // SOAPAction
                    &request,           // Input
                    &p_response );      // Output

    // Even if the transaction completes successfully, the SOAP body may have
    // been empty, so we need to test p_response.
    if( result.is_ok() && p_response )
    {
        std::wcout << L"Message: " << p_response->get_message() << L"\n";
        std::wcout << L"Host: " << p_response->get_hostName() << L"\n";
        std::cout << "Datetime: " << p_response->get_datetime() << "\n";
    }

    else
    {
        MK_LMX_DEBUG_ERROR_PTR( p_debug_error );       // Provides easier access to LMX debug information when debugging

        std::cout << "simple soap failed\n";
    }

    if( p_response )
        delete p_response;
}

void rest_example()
{
    // This example uses a RESTful service.  The response from the server
    // is received as a std::string and unmarshalled from that.

    std::cout << "\nIn rest_example()\n";

    lmx::c_winhttp http( "codalogic.com" );

    std::string output;

    int error = http.rest_query( "services/", std::string( "name=" ) + your_name, &output );

    if( error == 0 )
    {
        lmx::elmx_error lmx_error;
        const SOAPExample::c_Response response( output, &lmx_error );

        if( lmx_error == lmx::ELMX_OK )
        {
            std::wcout << L"Message: " << response.get_message() << L"\n";
            std::wcout << L"Host: " << response.get_hostName() << L"\n";
            std::cout << "Datetime: " << response.get_datetime() << "\n";
        }
    }
    else
    {
        std::cout << "Error code (" << error << ") occurred in rest_example()\n";
    }
}
